# Część 7: Interakcje z Klawiaturą i Myszą w Dynamicznych Programach Graficznych

Dla gier komputerowych, aplikacji graficznych oraz aplikacji mobilnych mysz stanowi kluczowy sposób interakcji z oknami graficznymi. W tym miejscu „mysz” obejmuje również dotyk palcem, ponieważ działa on w większości przypadków podobnie. Mysz jest wskaźnikiem poruszającym się po ekranie i reprezentuje pozycję piksela (zwykle na czubku strzałki, jeśli korzystamy z takiego kursora). Ta pozycja jest wyrażana jako współrzędna (x, y) w 2D, czyli w oknie graficznym.

Chociaż technicznie można byłoby tworzyć aplikacje graficzne bez interakcji, posiadanie tych urządzeń wejściowych okazuje się niezwykle przydatne przy testowaniu aplikacji, poruszaniu się w wirtualnym świecie i wchodzeniu w interakcję z modelami lub obiektami interfejsu użytkownika. Dlatego ten temat pojawia się już teraz, byśmy mogli eksplorować koncepcje graficzne w późniejszych częściach, mając do dyspozycji takie wejście. Ta wiedza będzie także przydatna przy tworzeniu własnych aplikacji. Poruszymy następujące kwestie:

- Korzystanie z interakcji myszy
- Dodawanie poleceń klawiatury

## Wymagania techniczne

W tej częsci będziemy korzystać z Pythona, PyCharm, Pygame i PyOpenGL, tak jak w poprzednich częściach.

**Uwaga:** Zakładamy, że masz przygotowany podstawowy szkielet aplikacji z poprzednich części zajęć – utworzone okno graficzne Pygame z kontekstem OpenGL, pętlę renderującą scenę oraz podstawowe klasy (np. klasa `Object` dla obiektów w scenie z listą komponentów, klasa `Transform` przechowująca pozycję obiektu w przestrzeni itp.).

W ramach tego ćwiczenia dodamy nowe pliki i dokonamy modyfikacji w kodzie, aby obsłużyć interakcję użytkownika:

---

## Korzystanie z interakcji myszy

Pozycja myszy na ekranie jest rejestrowana jako pozycja piksela w oparciu o układ współrzędnych ekranu (ang. screen coordinate system). Jak odkryłeś w poprzednich częściach, w domyślnym układzie współrzędnych ekranu punkt (0, 0) znajduje się w lewym górnym rogu okna graficznego.

Działania, jakie można wykonać myszą, obejmują:

- Naciśnięcie któregokolwiek przycisku myszy (lewego, prawego i środkowego)
- Zwolnienie nacisku przycisku myszy
- Przesunięcie myszy bez naciskania przycisku
- Przeciąganie (drag), czyli przesuwanie myszy przy wciśniętym przycisku
- Przewijanie (scroll), w którym użytkownik obraca rolkę myszy (lub na gładziku/ myszy dotykowej przesuwa palec w pionie lub poziomie)

Jak pokazano w części 1, jedną z metod przechwytywania zdarzeń myszy jest używanie fragmentu kodu takiego jak ten:

```python
for event in pygame.event.get():

    elif event.type == MOUSEBUTTONDOWN:
        # zrób coś
```

Pygame rozpoznaje następujące zdarzenia myszy, oprócz `MOUSEBUTTONDOWN`:

- `MOUSEBUTTONUP`
- `MOUSEMOTION`
- `MOUSEWHEEL`

Najlepszym sposobem, by je poznać, jest ich zaimplementowanie. Używaliśmy już `MOUSEBUTTONDOWN` w części 1 do rysowania linii i wielokątów w oknie, więc przyjrzyjmy się pozostałym.

### Obsługa myszy

1. Utwórz nowy plik Pythona o nazwie `MouseDrawing.py` i zacznij od następującego kodu:

   ```python
   import pygame
   from pygame.locals import *

   pygame.init()

   screen_width = 800
   screen_height = 800
   screen = pygame.display.set_mode((screen_width,
                                     screen_height))

   done = False
   white = pygame.Color(255, 255, 255)

   while not done:
       for event in pygame.event.get():
           if event.type == pygame.QUIT:
               done = True
           elif event.type == MOUSEBUTTONDOWN:
               pygame.draw.rect(screen, white,
                                (pygame.mouse.get_pos(),
                                (5, 5)))

       pygame.display.update()

   pygame.quit()
   ```

   Po uruchomieniu tego skryptu możesz rysować mały kwadrat w miejscu kliknięcia myszy, gdy naciśniesz przycisk myszy.

2. Zmodyfikuj kod tak, aby rysował kwadraty wyłącznie wtedy, gdy zostanie naciśnięty lewy przycisk myszy, korzystając z wartości `button == 1`:

   ```python
   elif event.type == MOUSEBUTTONDOWN:
       if event.button == 1:
           pygame.draw.rect(screen, white,
                            (pygame.mouse.get_pos(),
                             (5, 5)))
   ```

   W ten sposób sprawdzasz, który przycisk myszy został użyty. W tym przypadku sprawdzamy, czy to lewy przycisk (1).

3. Możesz też sprawdzać inne przyciski z użyciem wartości od 2 do 5:

   - 1: lewy przycisk
   - 2: środkowy przycisk
   - 3: prawy przycisk
   - 4: przewijanie w górę
   - 5: przewijanie w dół

4. Zmienimy teraz kod tak, aby rysował kwadraty przy przesuwaniu myszy z wciśniętym lewym przyciskiem. Zaktualizuj skrypt w następujący sposób:

   ```python
   mouse_down = False

   while not done:
       for event in pygame.event.get():
           if event.type == pygame.QUIT:
               done = True
           elif event.type == MOUSEBUTTONDOWN and \
                   event.button == 1:
               mouse_down = True
           elif event.type == MOUSEBUTTONUP and \
                   event.button == 1:
               mouse_down = False
           elif event.type == MOUSEMOTION and \
                  mouse_down is True:
               pygame.draw.rect(screen, white,
                                (pygame.mouse.get_pos(),
                                (5, 5)))

       pygame.display.update()
   ```

   W tym kodzie utworzono zmienną typu bool, aby śledzić stan przycisku myszy. Gdy przycisk zostaje wciśnięty, zmienna przyjmuje wartość `True`, a gdy przycisk zostanie zwolniony, wartość `False`. W momencie wykrycia ruchu myszy (`MOUSEMOTION`), jeśli `mouse_down == True`, rysujemy kwadrat w aktualnym położeniu wskaźnika myszy. Przy szybkim ruchu myszy mogą jednak pojawiać się luki między rysowanymi kwadratami.


5. Aby rysować ciągłą linię w oknie, zamiast pojedynczych kwadratów, musimy rysować odcinek między ostatnią a obecną pozycją myszy. To zapobiegnie lukom, jeśli mysz przesuwa się szybciej niż pętla główna nadąża z aktualizacją. Oznacza to, że musimy przechowywać pozycję myszy z poprzedniej iteracji pętli i narysować odcinek od tej pozycji do bieżącej. Można to zrobić tak:

   ```python
   last_mouse_pos = (0, 0)

   while not done:
       for event in pygame.event.get():
           if event.type == pygame.QUIT:
               done = True
           elif event.type == MOUSEBUTTONDOWN and \
                              event.button == 1:
               mouse_down = True
               last_mouse_pos = pygame.mouse.get_pos()
           elif event.type == MOUSEBUTTONUP and \
                              event.button == 1:
               mouse_down = False
           elif event.type == MOUSEMOTION and \
                mouse_down is True:
               pygame.draw.line(screen, white,
                                last_mouse_pos,
                                pygame.mouse.get_pos(),
                                5)
               last_mouse_pos = pygame.mouse.get_pos()

       pygame.display.update()
   ```

   Zamiast rysować kwadrat w punkcie pod myszą, używamy funkcji `pygame.draw.line()`. Przyjmuje ona parametr ekranu, kolor linii, punkt początkowy, punkt końcowy oraz grubość linii. Teraz możesz rysować bez przerw, nawet jeśli poruszysz myszą szybko.

Często, gdy używamy myszy w środowisku graficznym, użytkownik będzie chciał kliknąć np. przycisk lub obiekt. W tym celu należy obliczyć, czy pozycja myszy znajduje się w granicach wizualnego obszaru danego obiektu. W przypadku przycisku (czyli prostokąta) pozycja myszy musi znaleźć się w zakresie od lewego-górnego wierzchołka do prawego-dolnego wierzchołka przycisku:

![alt text](button.png)

Załóżmy, że nasz przycisk zaczyna się w punkcie `(x, y)`, a jego szerokość i wysokość to odpowiednio `(width, height)`. Warunek sprawdzający, czy mysz `(mx, my)` jest w obszarze przycisku, wygląda tak:

```python
if x < mx < (x + width) and y < my < (y + height):
```

### Ćwiczenie  
1. Stwórz zielony prostokąt w lewym górnym rogu okna, używając np. `green = pygame.Color(0, 255, 0)` i `button = (0, 0, 100, 30)`.  
2. Narysuj ten prostokąt w pętli głównej np. `pygame.draw.rect(screen, green, button)`.  
3. W tej samej pętli sprawdzaj, czy pozycja myszy znajduje się w obszarze prostokąta (używając `MOUSEMOTION`). Jeśli tak, wypisz w konsoli „Mouse Over”.

---

## Łączenie 2D i 3D w tym samym środowisku

W poprzedniej sekcji pracowaliśmy czysto w 2D z użyciem Pygame i rysowaliśmy np. prostokąty. Co jednak, jeśli chcemy połączyć środowisko 2D i 3D? Korzystając z OpenGL, mieszamy widok ortograficzny (2D) z widokiem perspektywicznym (3D). Można to zrobić, dzieląc obiekty do rysowania tak, aby część z nich była rysowana w trybie 3D, a część w trybie 2D. Najlepiej wyjaśnić to na przykładzie.

### Dodawanie przycisku

W tym ćwiczeniu nauczysz się, jak połączyć środowisko 2D i 3D w tym samym oknie. Postępuj zgodnie z tymi krokami:

1. Utwórz nowy plik Pythona `AddingButton.py` i dodaj do niego następujący kod:

   ```python
   from Object import *
   from Cube import *
   from pygame.locals import *
   from OpenGL.GL import *
   from OpenGL.GLU import *

   pygame.init()
   screen_width = 800
   screen_height = 600
   pygame.display.set_caption('OpenGL in Python')
   screen = pygame.display.set_mode((screen_width, screen_height),
                                    DOUBLEBUF | OPENGL)

   done = False
   white = pygame.Color(255, 255, 255)
   objects_3d = []
   objects_2d = []

   cube = Object("Cube")
   cube.add_component(Transform((0, 0, -5)))
   cube.add_component(Cube(GL_POLYGON, "plik_tekstury"))
   objects_3d.append(cube)

   clock = pygame.time.Clock()
   fps = 30

   def set_2d():
       glMatrixMode(GL_PROJECTION)
       glLoadIdentity()
       gluOrtho2D(0, screen.get_width(), 0,
                  screen.get_height())
       glMatrixMode(GL_MODELVIEW)
       glLoadIdentity()
       glViewport(0, 0, screen.get_width(),
                  screen.get_height())

   def set_3d():
       glMatrixMode(GL_PROJECTION)
       glLoadIdentity()
       gluPerspective(60, (screen_width / screen_height),
                      0.1, 100.0)
       glMatrixMode(GL_MODELVIEW)
       glLoadIdentity()
       glViewport(0, 0, screen.get_width(),
                  screen.get_height())
       glEnable(GL_DEPTH_TEST)

   while not done:
       for event in pygame.event.get():
           if event.type == pygame.QUIT:
               done = True

       glPushMatrix()
       glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

       set_3d()
       for o in objects_3d:
           o.update()

       set_2d()
       for o in objects_2d:
           o.update()

       glPopMatrix()
       pygame.display.flip()
       clock.tick(fps)

   pygame.quit()
   ```

   Aby to zadziałało, potrzebujesz także skryptów Pythona z poprzedniej części (`Cube.py`, `Object.py`, `Mesh3D.py`, `Transform.py`).  
   Zwróć uwagę na nowe metody `set_2d()` i `set_3d()`. Każda z nich ustawia macierze projekcji i widoku pod kątem rysowania obiektów w 2D bądź 3D. Obiekty 2D będą rysowane w trybie ortograficznym (`gluOrtho2D()`), a obiekty 3D – w trybie perspektywicznym (`gluPerspective()`).  

2. Czas na stworzenie przycisku. Utwórz nowy plik Pythona `Button.py` i wklej następujący kod:

   ```python
   from OpenGL.GL import *

   class Button:
       def __init__(self, screen, position, width,
                    height, color,
                    o_color, p_color):
           self.screen = screen
           self.position = position
           self.width = width
           self.height = height
           self.normal_color = color
           self.over_color = o_color
           self.pressed_color = p_color

       def draw(self):
           glPushMatrix()
           glLoadIdentity()
           glBegin(GL_POLYGON)
           glVertex2f(self.position[0], self.position[1])
           glVertex2f(self.position[0] + self.width,
                       self.position[1])
           glVertex2f(self.position[0] + self.width,
                       self.position[1] + self.height)
           glVertex2f(self.position[0],
                       self.position[1] + self.height)
           glEnd()
           glPopMatrix()
   ```

   Klasa `Button` określa przycisk jako prostokąt rysowany funkcją OpenGL `GL_POLYGON`.

3. Dodaj przycisk do `AddingButtons.py`:

   ```python
   from Button import *

   cube.add_component(Cube(GL_POLYGON,
                           "plik_tekstury"))
   objects_3d.append(cube)

   white = pygame.Color(255, 255, 255)
   green = pygame.Color(0, 255, 0)
   blue = pygame.Color(0, 0, 255)

   button1 = Object("Button")
   button1.add_component(Button(screen, (0, 0), 100, 50, white, green, blue))
   objects_2d.append(button1)

   clock = pygame.time.Clock()
   fps = 30
   ```

4. Dodatkowo zmodyfikuj `Objects.py` w taki sposób, aby umożliwić rysowanie przycisku:

   ```python
   def update(self, events=None):
       glPushMatrix()
       for c in self.components:
           if isinstance(c, Transform):
               pos = c.get_position()
               glTranslatef(pos.x, pos.y, pos.z)
           if isinstance(c, Mesh3D):
               c.draw()
           if isinstance(c, Button): # obsługa komponentu Button
               c.draw()
       glPopMatrix()
   ```

   Po uruchomieniu tego kodu ujrzysz kostkę pośrodku ekranu w widoku 3D oraz przycisk w lewym dolnym rogu, w widoku 2D.


Możesz zauważyć, że przycisk nie wyświetla się w oczekiwanym miejscu (np. myślałeś, że `(0,0)` to górny-lewy róg), a kolor przycisku jest inny niż zakładałeś (np. przycisk przejął teksturę kostki). Wynika to z użycia widoku ortograficznego i konieczności wyłączenia trybu texturowania po narysowaniu kostki 3D.

Aby to poprawić, w pliku `Mesh3D.py`, po zakończeniu rysowania kostki (w obrębie metody `draw()`), należy wyłączyć texturowanie:

```python
glDisable(GL_TEXTURE_2D)
```

i ustawić kolor przycisku, np.:

```python
glColor3f(self.normal_color[0], self.normal_color[1], self.normal_color[2])
```

dla stanu domyślnego (normalnego) przycisku.

---

## Konwertowanie myszy do współrzędnych projekcji

Jeśli w `gluOrtho2D()` zmienimy zakres np. na `gluOrtho2D(0, 1600, 0, 1200)`, nasz przycisk może wydać się mniejszy o połowę (względem poprzednich wartości), bo wirtualna „rozdzielczość” 2D została powiększona względem rzeczywistego rozmiaru okna. Mysz wciąż zwraca pozycję w układzie współrzędnych okna (np. 0..799, 0..599), więc musimy przeliczyć te wartości na nowy układ ortograficzny.

![alt text](screen_coords.png)

Załóżmy, że aktualnie mamy okno Pygame o wymiarach 800×600 (układ myszy 0..799, 0..599) oraz ortho2D 0..1600, 0..1200. Potrzebujemy funkcji, która przeniesie punkt `(mx, my)` z jednego zakresu do drugiego.

Można to wyrazić następującą zależnością proporcji, a następnie zaimplementować jako funkcję `map_value(current_min, current_max, new_min, new_max, value)`:

```python
def map_value(current_min, current_max, new_min, new_max,
              value):
    current_range = current_max - current_min
    new_range = new_max - new_min
    return new_min + new_range * ((value - current_min)/
                                  current_range)
```

W ten sposób łatwo zmapujesz np. `my` z przedziału `[0, 600]` na `[1200, 0]`, co odwraca układ współrzędnych w osi Y. Funkcję mapującą umieść w nowym pliku **Utils.py**.

Aktualizacja klasy Button z uwzględnieniem mapowania współrzędnych myszy:
```python
from OpenGL.GL import *
from Utils import *
class Button:
..
    def draw(self, events):
        mouse_pos = pygame.mouse.get_pos()
        mx = map_value(0, 800, 0, 1600, mouse_pos[0])
        my = map_value(0, 600, 1200, 0, mouse_pos[1])
        glPushMatrix()
        glLoadIdentity()
        # jeśli kursor nad przyciskiem
        if self.position[0] < mx < (self.position[0] + self.width) and  self.position[1] < my < (self.position[1] + self.height):
            glColor3f(self.over_color[0], self.over_color[1], self.over_color[2])
        else:
            glColor3f(self.normal_color[0], self.normal_color[1], self.normal_color[2])
        glBegin(GL_POLYGON)
```

Aktualizacja w klasie Object (przekazanie zdarzeń do metody `draw()`):
```python
def update(self, events = None):
    glPushMatrix()
    for c in self.components:
        ..
        if isinstance(c, Button):
            c.draw(events)
    glPopMatrix()
```

Aktualizacja w skrypcie AddingButtons.py (przechwycenie zdarzeń w pętli głównej i przekazanie do metody `update()`):
```python
while not done:
    events = pygame.event.get()
    for event in events:
        if event.type == pygame.QUIT:
        done = True
    ..
    set_3d()
    for o in objects_3d:
        o.update(events)

    set_2d()
    for o in objects_2d:
        o.update(events)

```
---

### Ćwiczenia

1. W aplikacji do rysowania (z poprzednich kroków) sprawdzaliśmy kliknięcie myszy, aby rysować. W klasie `Button` istnieje właściwość `pressed_color`, której używamy do koloru przycisku w momencie kliknięcia. Zaimplementuj logikę tak, by przycisk faktycznie przyjmował kolor `pressed_color` po naciśnięciu (lewym przyciskiem), a po zwolnieniu wracał do koloru `normal_color` lub `over_color` w zależności od tego, czy kursor wciąż się nad nim znajduje.

2. Obecnie w kodzie znajdują się zapisane na sztywno wartości szerokości i wysokości okna, np. 800 i 600 oraz 1600 i 1200. Stwórz osobny plik `Settings.py`, w którym zadeklarujesz te wymiary (np. `window_dimensions` i `gui_dimensions`), a następnie odwołuj się do nich w całym projekcie. Ułatwi to modyfikacje rozmiarów w przyszłości i będzie dobrą praktyką programistyczną.

---

## Dodawanie akcji kliknięcia przycisku

W późniejszych częściach podłączymy do przycisków bardziej rozbudowane funkcje. Na razie jednak pokażemy, jak wywołać prostą funkcję przy kliknięciu (funkcję zdefiniuj a pliku AddingButtons.py):

```python
def button_click():
    print("Hello Button")
```

Następnie, przy tworzeniu przycisku, przekaż tę funkcję jako argument konstruktorowi klasy `Button`. W `Button` przechowamy ją w zmiennej np. `on_click` i wywołamy w `MOUSEBUTTONDOWN`:

```python
button1.add_component(Button(screen, (0, 0), 100, 50,
                             green, white, blue,
                             button_click))
```

A w klasie `Button`:

```python
def __init__(... on_click):
    ...
    self.on_click = on_click
    ...
```

Obsługa zdarzeń w metodzie `draw()` klasy Button:
```python 
...
for e in events:
    if e.type == MOUSEBUTTONDOWN and e.button == 1:
        self.mouse_down = True
        self.on_click() # wywołanie przekazanej funkcji
    elif e.type == MOUSEBUTTONUP and e.button == 1:
        self.mouse_down = False
...
```

Po uruchomieniu skryptu, przy każdym kliknięciu przycisku zobaczysz w konsoli „Hello Button”.

---

## Dodawanie poleceń klawiatury

Skoro poznaliśmy podstawy interakcji myszy, klawiatura wyda się prostsza: posiada mniej zdarzeń (przede wszystkim `KEYDOWN` i `KEYUP`). Można też łączyć naciśnięcie wielu klawiszy jednocześnie. Najczęściej klawisze służą do sterowania obiektami w scenie, np. strzałki lub WASD do poruszania, spacja do skakania/strzału itd.

### Przesuwanie obiektu w 3D za pomocą klawiszy

1. W klasie `Transform` (z poprzedniej części) dodajmy metodę, która pozwoli przesunąć obiekt po osi X:

   ```python
   def move_x(self, amount):
       self.position = pygame.math.Vector3(
           self.position.x + amount,
           self.position.y,
           self.position.z
       )
   ```

2. Gdy klawisz strzałki w lewo zostanie wciśnięty, wywołujmy `move_x()` w `AddingButtons.py`:

   ```python
   while not done:
       events = pygame.event.get()
       for event in events:
           if event.type == pygame.QUIT:
               done = True
           elif event.type == pygame.KEYDOWN:
               if event.key == pygame.K_LEFT:
                   trans: Transform = cube.get_component(Transform)
                   if trans is not None:
                       trans.move_x(-0.1) #przesuwamy
       ...
   ```

3. Dodajmy funkcję `get_component` do klasy `Object`:

   ```python
   def get_component(self, class_type):
       for c in self.components:
           if type(c) is class_type:
               return c
       return None
   ```

### Ćwiczenia

1. Dodaj obsługę przesunięcia w prawo, gdy naciśnięty zostanie klawisz `K_RIGHT`.  
2. Dopisz w klasie `Transform` metodę `move_y`, a następnie obsłuż strzałki w górę i w dół (`K_UP`, `K_DOWN`), by przesuwać kostkę w osi Y.

Wielu graczy preferuje przytrzymanie klawisza, by obiekt przesuwał się do momentu zwolnienia klawisza (zamiast jednorazowego kliknięcia). Można to osiągnąć podobnie jak w przypadku rysowania, gdy mysz jest wciśnięta. Alternatywnie możemy użyć wbudowanej w Pygame funkcji `get_pressed()`, która zwraca słownik ze stanem wszystkich klawiszy w danym momencie.

```python
keys = pygame.key.get_pressed()
if keys[pygame.K_LEFT]:
    trans.move_x(-0.1)
if keys[pygame.K_RIGHT]:
    trans.move_x(0.1)
```

---

## Podsumowanie

W tej części skupiliśmy się na obsłudze myszy i klawiatury w aplikacjach graficznych. Omówiliśmy:

- Metody przechwytywania zdarzeń myszy: naciśnięcie, zwolnienie, ruch, przeciąganie (drag), przewijanie (scroll).
- Wykorzystywanie tych zdarzeń do rysowania (np. linii) oraz interakcji z przyciskami 2D.
- Łączenie widoku ortograficznego do rysowania interfejsu z widokiem perspektywicznym do renderowania obiektów 3D w tej samej aplikacji.
- Konieczność mapowania współrzędnych myszy okna do układu rzutowania ortograficznego.
- Obsługę zdarzeń klawiatury – zarówno pojedynczych (`KEYDOWN`), jak i ciągłych (`get_pressed()`).
